import contextlib
import csv
import gc
import glob
import importlib
import json
import os
import shutil
import sys
import traceback
import warnings
from datetime import datetime
from pathlib import Path
from stat import S_IREAD, S_IWRITE
import tarfile
import tempfile
from typing import List, Optional, Any
from urllib.request import urlretrieve

from semantic_version import Version, SimpleSpec

import cadetrdm
from cadetrdm import Options
from cadetrdm.io_utils import delete_path, test_for_lfs
from cadetrdm.io_utils import recursive_chmod, write_lines_to_file, wait_for_user, init_lfs
from cadetrdm.jupyter_functionality import Notebook
from cadetrdm.logging import OutputLog, LogEntry
from cadetrdm.remote_integration import GitHubRemote, GitLabRemote
from cadetrdm.web_utils import ssh_url_to_http_url

try:
    import git
except ImportError:
    # Adding this hint to save users the confusion of trying $pip install git
    raise ImportError("No module named git, please install the gitpython package")


def validate_is_output_repo(path_to_repo):
    with open(os.path.join(path_to_repo, ".cadet-rdm-data.json"), "r", encoding="utf-8") as file_handle:
        rdm_data = json.load(file_handle)
        if rdm_data["is_project_repo"]:
            raise ValueError("Please use the URL to the output repository.")


class GitRepo:
    def __init__(self, path=None, search_parent_directories=True, *args, **kwargs):
        """
        Base class handling most git workflows.

        :param path:
            Path to the root directory of the repository.
        :param search_parent_directories:
            if True, all parent directories will be searched for a valid repo as well.

            Please note that this was the default behaviour in older versions of GitPython,
            which is considered a bug though.
        :param args:
            Args handed to git.Repo()
        :param kwargs:
            Kwargs handed to git.Repo()
        """
        if path is None or path == ".":
            path = os.getcwd()

        if type(path) is str:
            path = Path(path)

        self._git_repo = git.Repo(path, search_parent_directories=search_parent_directories, *args, **kwargs)
        self._git = self._git_repo.git

        self._most_recent_branch = self.active_branch.name
        self._earliest_commit = None

        existing_branches = [branch.name for branch in self._git_repo.branches]
        if len(existing_branches) != 0 and "main" not in existing_branches and "master" in existing_branches:
            self.main_branch = "master"
        else:
            self.main_branch = "main"

        self.add = self._git.add

    def __enter__(self) -> "Repo":
        return self

    def __exit__(self, *args: Any) -> None:
        self._git_repo.close()

    def __del__(self) -> None:
        try:
            self._git_repo.close()
        except Exception:
            traceback.print_exc()
            pass

    @property
    def active_branch(self):
        try:
            active_branch = self._git_repo.active_branch
            return active_branch
        except TypeError as e:
            if "HEAD is a detached" in str(e):
                class DetachedHeadBranch:
                    name = "detached_head"

                    def __repr__(self):
                        return "detached_head"

                    def __str__(self):
                        return "detached_head"

                return DetachedHeadBranch()
            else:
                raise e

    @property
    def untracked_files(self):
        return self._git_repo.untracked_files

    @property
    def current_commit_hash(self):
        return str(self.head.commit)

    @property
    def path(self):
        return Path(self._git_repo.working_dir)

    @property
    def bare(self):
        return self._git_repo.bare

    @property
    def working_dir(self):
        print("Deprecation Warning. .working_dir is getting replaced with .path")
        return Path(self._git_repo.working_dir)

    @property
    def head(self):
        return self._git_repo.head

    @property
    def remotes(self):
        return self._git_repo.remotes

    @property
    def remote_urls(self):
        if len(self.remotes) == 0:
            print(RuntimeWarning(f"No remote for repo at {self.path} set yet. Please add remote ASAP."))
        return [str(remote.url) for remote in self.remotes]

    @property
    def url(self):
        return self.remote_urls[0]

    @property
    def earliest_commit(self):
        if self._earliest_commit is None:
            *_, earliest_commit = self._git_repo.iter_commits()
            self._earliest_commit = earliest_commit
        return self._earliest_commit

    @property
    def tags(self):
        return list()

    @property
    def data_json_path(self):
        return self.path / ".cadet-rdm-data.json"

    @property
    def cache_json_path(self):
        return self.path / ".cadet-rdm-cache.json"

    @property
    def has_changes_upstream(self):
        try:
            remote = self.remotes[0]
            remote_branches = remote.fetch()
            correct_remote_branches = [fetch_info for fetch_info in remote_branches
                                       if fetch_info.name == f"{remote.name}/{self.active_branch.name}"]
            if len(correct_remote_branches) > 1:
                raise RuntimeError(f"Remote has multiple branches matching local branch {self.active_branch.name}: "
                                   f"{[branch.name for branch in correct_remote_branches]}")
            remote_hash = str(correct_remote_branches[0].commit)

            if self.current_commit_hash != remote_hash and remote_hash not in self.log:
                return True
            elif self.current_commit_hash != remote_hash and remote_hash in self.log:
                print("Local repository is ahead of remote. This could be due to CADET-RDM version updates.")
                return False
            else:
                return False

        except git.GitCommandError as e:
            traceback.print_exc()
            print(f"Git command error in {self.path}: {e}")

    def fetch(self):
        self._git.fetch()

    def update(self):
        try:
            self.fetch()

            if self.has_changes_upstream:
                print(f"New changes detected in {self.remotes[0]}, pulling updates...")
                self.remotes[0].pull()
                self._git.reset(["--hard", f"{self.remotes[0]}/{self.active_branch}"])

        except git.GitCommandError as e:
            traceback.print_exc()
            print(f"Git command error in {self.path}: {e}")

    @classmethod
    def clone(cls, url: str, to_path: str | Path = None, multi_options: Optional[List[str]] = None, **kwargs):
        """
        Clone a remote repository

        :param url:
        :param to_path:
        :param multi_options: A list of Clone options that can be provided multiple times.
            One option per list item which is passed exactly as specified to clone.
            For example: ['--config core.filemode=false', '--config core.ignorecase',
            '--recurse-submodule=repo1_path', '--recurse-submodule=repo2_path']
        :return:
        """

        # prevent git terminal prompts from interrupting the process.
        previous_environment_variables = cls._git_environ_setup()

        if to_path is None:
            to_path = url.split("/")[-1].replace(".git", "")
        print(f"Cloning {url} into {to_path}")

        try:
            git.Repo.clone_from(url, to_path, multi_options=multi_options, **kwargs)
        except git.exc.GitCommandError as e:
            print(f"Clone from {url} failed with {e, e.stderr, e.stdout}.")
            print(f"Retrying with {ssh_url_to_http_url(url)}.")
            try:
                git.Repo.clone_from(ssh_url_to_http_url(url), to_path, multi_options=multi_options, **kwargs)
            except Exception as e_inner:
                print(f"Clone from {ssh_url_to_http_url(url)} failed with: ")
                traceback.print_exc()
                raise e_inner
        finally:
            cls._git_environ_reset(previous_environment_variables)

        instance = cls(to_path)
        return instance

    @staticmethod
    def _git_environ_setup():
        environment_variables = {
            "GIT_TERMINAL_PROMPT": "0",
            "GIT_SSH_COMMAND": "ssh -o StrictHostKeyChecking=yes",
            "GIT_CLONE_PROTECTION_ACTIVE": "0",
        }
        previous_environment_variables = {key: None for key in environment_variables.keys()}
        for key, value in environment_variables.items():
            if key in os.environ:
                previous_environment_variables[key] = os.environ[key]
            os.environ[key] = value

        return previous_environment_variables

    @staticmethod
    def _git_environ_reset(previous_environment_variables):
        for key, previous_value in previous_environment_variables.items():
            if previous_value is None:
                os.environ.pop(key)
            else:
                os.environ[key] = previous_value

    def checkout(self, *args, **kwargs):
        self._most_recent_branch = self.active_branch
        self._git.checkout(*args, **kwargs)

    def remote_set_url(self, name: str, url: str):
        """
        Set the url of a named remote.

        :param name:
        :param url:
        """
        self._git_repo.remotes[name].set_url(url)

    def commit(self, message: str | None = None, add_all=True, verbosity=1):
        """
        Commit current state of the repository.

        :param message:
            Commit message
        :param add_all:
            Option to add all changed and new files to git automatically.
        :param verbosity:
            Option to choose degree of printed feedback.
        """

        if not self.has_uncomitted_changes:
            if verbosity >= 1:
                print(f"No changes to commit in repo {self.path}")
            return

        print("Found changes in these files:")
        for file in [self.untracked_files + self.changed_files]:
            print(" -", file)

        if message is None:
            message = input("Please enter a commit message for these code changes or 'N' to cancel.\n")
            if message.upper().replace(" ", "") == "N" or message.upper().replace(" ", "") == "":
                raise KeyboardInterrupt

        print(f"Commiting changes to repo {self.path}")
        if add_all:
            self.add(".")

        try:
            commit_return = self._git.commit("-m", message)
            if verbosity >= 1:
                print("\n" + commit_return + "\n")
        except:
            pass

    def git_ammend(self, ):
        """
        Call git commit with options --amend --no-edit
        """
        self._git.commit("--amend", "--no-edit")

    @property
    def status(self):
        return self._git.status()

    @property
    def log(self):
        return self._git.log()

    def log_oneline(self):
        return self._git.log("--oneline")

    def print_status(self):
        """
        prints git status
        """
        print(self._git.status())

    def print_log(self):
        """
        Prints the git log
        """
        print(self._git.log())

    def stash_all_changes(self):
        """
        Adds all untracked files to git and then stashes all changes.
        Will raise a RuntimeError if no changes are found.
        """
        if not self.has_uncomitted_changes:
            warnings.warn("No changes in repo to stash.")
            return
        self.add(".")
        self._git.stash()

    def apply_stashed_changes(self):
        """
        Apply the last stashed changes.
        If a "CONFLICT (modify/delete)" error is encountered, this is ignored.
        All other errors are raised.
        """
        try:
            self._git.stash('pop')  # equivalent to $ git stash pop
        except git.exc.GitCommandError as e:
            # Will raise error because the stash cannot be applied without conflicts. This is expected
            if 'CONFLICT (modify/delete)' in e.stdout:
                pass
            else:
                raise e

    def test_for_uncommitted_changes(self):
        """
        Raise a RuntimeError if uncommitted changes are in the repository.
        :return:
        """
        if self.has_uncomitted_changes:
            raise RuntimeError(f"Found uncommitted changes in the repository {self.path}.")

    def push(self, remote=None, local_branch=None, remote_branch=None, push_all=True):
        """
        Push local branch to remote.

        :param remote:
            Name of the remote to push to.
        :param local_branch:
            Name of the local branch to push.
        :param remote_branch:
            Name of the remote branch to push to.
        :return:
        """
        if local_branch is None:
            local_branch = self.active_branch
        if remote_branch is None:
            remote_branch = local_branch
        if remote is None:
            if len(self._git_repo.remotes) == 0:
                raise RuntimeError("No remote has been set for this repository yet.")
            remote_list = [str(remote.name) for remote in self._git_repo.remotes]
        else:
            remote_list = [remote]

        if local_branch == self.main_branch or push_all:
            if push_all:
                self.checkout(self.main_branch)
            for remote in remote_list:
                remote_interface = self._git_repo.remotes[remote]
                try:
                    remote_interface.pull()
                except Exception as e:
                    print("Pulling from this remote failed with the following error:")
                    print(e)

        for remote in remote_list:
            remote_interface = self._git_repo.remotes[remote]

            if push_all:
                push_results = remote_interface.push(all=True)
            else:
                push_results = remote_interface.push(refspec=f'{local_branch}:{remote_branch}')

            for push_res in push_results:
                print(push_res.summary)

        if hasattr(self, "output_repo") and push_all:
            self.output_repo.push()

    def delete_active_branch_if_branch_is_empty(self):
        """
        Delete the currently active branch and checkout the main branch
        :return:
        """
        previous_branch = self.active_branch.name
        if previous_branch == self.main_branch:
            return

        commit_of_current_main = str(self._git.rev_parse(self.main_branch))
        commit_of_current_branch = str(self.head.commit)
        if commit_of_current_branch == commit_of_current_main:
            print("Removing empty branch", previous_branch)
            self._git.checkout(self.main_branch)
            self._git.branch("-d", previous_branch)

    def add_all_files(self, automatically_add_new_files=True):
        """
        Stage all changes to git. This includes new, untracked files as well as modified files.
        :param automatically_add_new_files:
            If this is set to false a user input will be prompted if untracked files are about to be added.
        :return:
            List of all staged changes.
        """
        self.add(".")

    def _reset_hard_to_head(self, force_entry=False):
        if not force_entry:
            proceed = wait_for_user(f'The output directory contains the following uncommitted changes:\n'
                                    f'{self.untracked_files + self.changed_files}\n'
                                    f' These will be lost if you continue\n'
                                    f'Proceed?')
        else:
            proceed = True
        if not proceed:
            raise KeyboardInterrupt
        # reset all tracked files to previous commit, -q silences output
        self._git.reset("-q", "--hard", "HEAD")
        # remove all untracked files and directories, -q silences output
        try:
            self._git.clean("-q", "-f", "-d")
        except git.exc.GitCommandError:
            recursive_chmod(self.path, S_IWRITE)
            self._git.clean("-q", "-f", "-d")

    @property
    def changed_files(self):
        changed_files = self._git.diff(None, name_only=True).split('\n')
        if "" in changed_files:
            changed_files.remove("")
        return changed_files

    @property
    def exist_uncomitted_changes(self):
        warnings.warn(
            "ProjectRepo.exist_uncomitted_changes will be removed in a future version. Please use .has_uncomitted_changes"
        )
        return self.has_uncomitted_changes

    @property
    def has_uncomitted_changes(self):
        return len(self._git.status("--porcelain")) > 0

    def ensure_relative_path(self, input_path):
        """
        Turn the input path into a relative path, relative to the repo working directory.

        :param input_path:
        :return:
        """
        if type(input_path) is str:
            input_path = Path(input_path)

        if input_path.is_absolute():
            relative_path = input_path.relative_to(self.path)
        else:
            relative_path = input_path
        return relative_path


class BaseRepo(GitRepo):
    def __init__(self, path=None, search_parent_directories=True, *args, **kwargs):
        """
        Base class handling most git workflows.

        :param path:
            Path to the root directory of the repository.
        :param search_parent_directories:
            if True, all parent directories will be searched for a valid repo as well.

            Please note that this was the default behaviour in older versions of GitPython,
            which is considered a bug though.
        :param args:
            Args handed to git.Repo()
        :param kwargs:
            Kwargs handed to git.Repo()
        """
        super().__init__(path, search_parent_directories, *args, **kwargs)
        self._metadata = self.load_metadata()

    def load_metadata(self):
        with open(self.data_json_path, "r", encoding="utf-8") as handle:
            metadata = json.load(handle)
        if "output_remotes" not in metadata and metadata["is_project_repo"]:
            # this enables upgrades from v0.0.23 to v0.0.24.
            output_remotes_path = self.path / "output_remotes.json"
            with open(output_remotes_path, "r", encoding="utf-8") as handle:
                output_remotes = json.load(handle)
            metadata["output_remotes"] = output_remotes
        return metadata

    def add_remote(self, remote_url, remote_name=None):
        """
        Add a remote to the repository.

        :param remote_url:
        :param remote_name:
        :return:
        """
        if remote_name is None:
            remote_name = "origin"
        self._git_repo.create_remote(remote_name, url=remote_url)
        if self._metadata["is_project_repo"]:
            # This folder is a project repo. Use a project repo class to easily access the output repo.
            output_repo = ProjectRepo(self.path).output_repo

            if output_repo.active_branch != output_repo.main_branch:
                if output_repo.has_uncomitted_changes:
                    output_repo.stash_all_changes()
                output_repo.checkout(output_repo.main_branch)
            output_repo.add_list_of_remotes_in_readme_file("Link to Project Repository", self.remote_urls)
            output_repo.add("README.md")
            output_repo.commit("Add remote for project repo", verbosity=0, add_all=False)
        if self._metadata["is_output_repo"]:
            # This folder is an output repo
            project_repo = ProjectRepo(self.path.parent)
            project_repo.update_output_remotes_json()
            project_repo.add_list_of_remotes_in_readme_file("Link to Output Repository", self.remote_urls)
            project_repo.add(project_repo.data_json_path)
            project_repo.add("README.md")
            project_repo.commit("Add remote for output repo", verbosity=0, add_all=False)

    def import_remote_repo(self, source_repo_location, source_repo_branch, target_repo_location=None):
        """
        Import a remote repo and update the cadet-rdm-cache

        :param source_repo_location:
        Path or URL to the source repo.
        Example https://jugit.fz-juelich.de/IBG-1/ModSim/cadet/agile_cadet_rdm_presentation_output.git
        or git@jugit.fz-juelich.de:IBG-1/ModSim/cadet/agile_cadet_rdm_presentation_output.git

        :param source_repo_branch:
        Branch of the source repo to check out.

        :param target_repo_location:
        Place to store the repo. If None, the external_cache folder is used.

        :return:
        Path to the cloned repository
        """
        if "://" in str(source_repo_location):
            source_repo_name = source_repo_location.split("/")[-1]
        else:
            source_repo_name = Path(source_repo_location).name
        if target_repo_location is None:
            target_repo_location = self.path / "external_cache" / source_repo_name
        else:
            target_repo_location = self.path / target_repo_location

        self.add_path_to_gitignore(target_repo_location)

        print(f"Cloning from {source_repo_location} into {target_repo_location}")
        multi_options = ["--filter=blob:none", "--branch", source_repo_branch, "--single-branch"]
        repo = GitRepo.clone(url=source_repo_location, to_path=target_repo_location,
                             multi_options=multi_options)
        repo._git.clear_cache()
        repo._git_repo.close()

        self.update_cadet_rdm_cache_json(source_repo_branch=source_repo_branch,
                                         target_repo_location=target_repo_location,
                                         source_repo_location=source_repo_location)
        return target_repo_location

    def add_path_to_gitignore(self, path_to_be_ignored):
        """
        Add the path to the .gitignore file

        :param path_to_be_ignored:
        :return:
        """
        path_to_be_ignored = self.ensure_relative_path(path_to_be_ignored)
        with open(self.path / ".gitignore", "r", encoding="utf-8") as file_handle:
            gitignore = file_handle.readlines()
            gitignore[-1] += "\n"  # Sometimes there is no trailing newline
        if str(path_to_be_ignored) + "\n" not in gitignore:
            gitignore.append(str(path_to_be_ignored) + "\n")
        with open(self.path / ".gitignore", "w", encoding="utf-8") as file_handle:
            file_handle.writelines(gitignore)

    def update_cadet_rdm_cache_json(self, source_repo_location, source_repo_branch, target_repo_location):
        """
        Update the information in the .cadet_rdm_cache.json file

        :param source_repo_location:
        Path or URL to the source repo.
        :param source_repo_branch:
        Name of the branch to check out.
        :param target_repo_location:
        Path where to put the repo or data
        """
        if not self.cache_json_path.exists():
            with open(self.cache_json_path, "w", encoding="utf-8") as file_handle:
                file_handle.writelines("{}")

        with open(self.cache_json_path, "r", encoding="utf-8") as file_handle:
            rdm_cache = json.load(file_handle)

        repo = GitRepo(target_repo_location)
        commit_hash = repo.current_commit_hash
        if "__example/path/to/repo__" in rdm_cache.keys():
            rdm_cache.pop("__example/path/to/repo__")

        target_repo_location = str(self.ensure_relative_path(target_repo_location))

        if isinstance(source_repo_location, Path):
            source_repo_location = source_repo_location.as_posix()

        rdm_cache[target_repo_location] = {
            "source_repo_location": source_repo_location,
            "branch_name": source_repo_branch,
            "commit_hash": commit_hash,
        }

        with open(self.cache_json_path, "w", encoding="utf-8") as file_handle:
            json.dump(rdm_cache, file_handle, indent=2)

    def verify_unchanged_cache(self):
        """
        Verify that all repos referenced in .cadet-rdm-data.json are
        in an unmodified state. Raises a RuntimeError if the commit hash has changed or if
        uncommited changes are found.

        :return:
        """

        with open(self.cache_json_path, "r", encoding="utf-8") as file_handle:
            rdm_cache = json.load(file_handle)

        if "__example/path/to/repo__" in rdm_cache.keys():
            rdm_cache.pop("__example/path/to/repo__")

        for repo_location, repo_info in rdm_cache.items():
            try:
                repo = GitRepo(repo_location)
                repo._git.clear_cache()
            except git.exc.NoSuchPathError:
                raise git.exc.NoSuchPathError(f"The imported repository at {repo_location} was not found.")

            self.verify_cache_folder_is_unchanged(repo_location, repo_info["commit_hash"])

    def verify_cache_folder_is_unchanged(self, repo_location, commit_hash):
        """
        Verify that the repo located at repo_location has no uncommited changes and that the current commit_hash
        is equal to the given commit_hash

        :param repo_location:
        :param commit_hash:
        :return:
        """
        repo = GitRepo(repo_location)
        commit_changed = repo.current_commit_hash != commit_hash
        uncommited_changes = repo.has_uncomitted_changes
        if commit_changed or uncommited_changes:
            raise RuntimeError(f"The contents of {repo_location} have been modified. Don't do that.")
        repo._git.clear_cache()

    def dump_package_list(self, target_folder):
        """
        Use "conda env export" and "pip freeze" to create environment.yml and pip_requirements.txt files.
        """
        if target_folder is not None:
            dump_path = target_folder
        else:
            dump_path = self.path
        print("Dumping conda environment.yml, this might take a moment.")
        try:
            os.system(f"conda env export > {dump_path}/conda_environment.yml")
            print("Dumping conda independent environment.yml, this might take a moment.")
            os.system(f"conda env export --from-history > {dump_path}/conda_independent_environment.yml")
        except Exception as e:
            print("Could not dump conda environment due to the following error:")
            print(e)
        print("Dumping pip requirements.txt.")
        os.system(f"pip freeze > {dump_path}/pip_requirements.txt")
        print("Dumping pip independent requirements.txt.")
        os.system(f"pip list --not-required --format freeze > {dump_path}/pip_independent_requirements.txt")

    def add_list_of_remotes_in_readme_file(self, repo_identifier: str, remotes_url_list: list):
        if len(remotes_url_list) > 0:
            remotes_url_list_http = [ssh_url_to_http_url(remote)
                                     for remote in remotes_url_list]
            output_link_line = " and ".join(f"[{repo_identifier}]({output_repo_remote})"
                                            for output_repo_remote in remotes_url_list_http) + "\n"

            readme_filepath = self.path / "README.md"
            with open(readme_filepath, "r", encoding="utf-8") as file_handle:
                filelines = file_handle.readlines()
                filelines_giving_output_repo = [i for i in range(len(filelines))
                                                if filelines[i].strip().startswith(f"[{repo_identifier}](")]
                if len(filelines_giving_output_repo) == 1:
                    line_to_be_modified = filelines_giving_output_repo[0]
                    filelines[line_to_be_modified] = output_link_line
                elif len(filelines_giving_output_repo) == 0:
                    filelines.append("The output repository can be found at:\n") #method can be used for project and output repositories, "the corresponding repository can be found at... would be better"
                    filelines.append(output_link_line)
                else:
                    raise RuntimeError(f"Multiple lines in the README.md at {readme_filepath}"
                                       f" link to the {repo_identifier}. "
                                       "Can't automatically update the link.")

            with open(readme_filepath, "w", encoding="utf-8") as file_handle:
                file_handle.writelines(filelines)
            self.add(readme_filepath)


class ProjectRepo(BaseRepo):
    def __init__(self, path=None, output_folder=None,
                 search_parent_directories=True, suppress_lfs_warning=False,
                 url=None, branch=None, options=None,
                 *args, **kwargs):
        """
        Class for Project-Repositories. Handles interaction between the project repo and
        the output (i.e. results) repo.

        :param path:
            Path to the root of the git repository.
        :param output_folder:
            Deprecated: Path to the root of the output repository.
        :param search_parent_directories:
            if True, all parent directories will be searched for a valid repo as well.

            Please note that this was the default behaviour in older versions of GitPython,
            which is considered a bug though.
        :param suppress_lfs_warning:
            Option to not test for git-lfs installation. Used if running a study in a Docker container
            from a system without git-lfs
        :param branch:
            Optional branch to check out upon initialization
        :param options:
            Options dictionary containing ...
        :param args:
            Additional args to be handed to BaseRepo.
        :param kwargs:
            Additional kwargs to be handed to BaseRepo.
        """
        if path is not None and not Path(path).exists():
            if url is None:
                raise ValueError(f"Could not find repository at path {path} and no url was given.")
            ProjectRepo.clone(url=url, to_path=path)

        super().__init__(path, search_parent_directories=search_parent_directories, *args, **kwargs)

        if not suppress_lfs_warning:
            test_for_lfs()

        if output_folder is not None:
            print("Deprecation Warning. Setting the outputfolder manually during repo instantiation is deprecated"
                  " and will be removed in a future update.")

        if not self.data_json_path.exists():
            raise RuntimeError(f"Folder {self.path} does not appear to be a CADET-RDM repository.")

        self._project_uuid = self._metadata["project_uuid"]
        self._output_uuid = self._metadata["output_uuid"]
        self._output_folder = self._metadata["output_remotes"]["output_folder_name"]
        self.options = options
        if not (self.path / self._output_folder).exists():
            print("Output repository was missing, cloning now.")
            self._clone_output_repo()
        self.output_repo = OutputRepo(self.path / self._output_folder)

        if self._metadata["cadet_rdm_version"] != cadetrdm.__version__:
            self._update_version(self._metadata, cadetrdm.__version__)

        self._on_context_enter_commit_hash = None
        self._is_in_context_manager = False
        self.options_hash = None

        if branch is not None:
            self.checkout(branch)

    @property
    def name(self):
        return self.path.parts[-1]

    @property
    def module(self):
        cur_dir = os.getcwd()

        os.chdir(self.path)
        sys.path.append(str(self.path))
        module = importlib.import_module(self.name)

        sys.path.remove(str(self.path))
        os.chdir(cur_dir)
        return module

    def _update_version(self, metadata, cadetrdm_version):
        current_version = Version.coerce(metadata["cadet_rdm_version"])

        changes_were_made = False

        if SimpleSpec("<0.0.9").match(current_version):
            changes_were_made = True
            self._convert_csv_to_tsv_if_necessary()
            self._add_jupytext_file(self.path)
        if SimpleSpec("<0.0.24").match(current_version):
            changes_were_made = True
            self._expand_tsv_header()
            output_remotes_path = self.path / "output_remotes.json"
            delete_path(output_remotes_path)
            self.add(output_remotes_path)
        if SimpleSpec("<0.0.34").match(current_version):
            changes_were_made = True
            self.fix_gitattributes_log_tsv()
        if SimpleSpec("<0.1.7").match(current_version):
            changes_were_made = True
            if self.output_repo.output_log.n_entries > 0:
                warnings.warn(
                    "Repo version has outdated options hashes. "
                    "Updating option hashes in output log.tsv."
                )
                self.output_repo.update_log_hashes()

        if changes_were_made:
            print(f"Repo version {metadata['cadet_rdm_version']} was outdated. "
                  f"Current CADET-RDM version is {cadetrdm.__version__}.\n Repo has been updated")
            metadata["cadet_rdm_version"] = cadetrdm_version
            with open(self.data_json_path, "w", encoding="utf-8") as f:
                json.dump(metadata, f, indent=2)
            self.add(self.data_json_path)
            self.commit("update cadetrdm version", add_all=False)

    def fix_gitattributes_log_tsv(self):
        file = self.output_path / ".gitattributes"
        with open(file, encoding="utf-8") as handle:
            lines = handle.readlines()
        lines = [line.replace("rdm-log.tsv", "log.tsv") for line in lines]
        with open(file, "w", encoding="utf-8") as handle:
            handle.writelines(lines)
        self.output_repo.add(".gitattributes")
        self.output_repo.commit("Update gitattributes")

    def _clone_output_repo(self, multi_options: List[str] = None):
        metadata = self.load_metadata()
        output_remotes = metadata["output_remotes"]
        output_path = self.path / output_remotes["output_folder_name"]
        ssh_remotes = list(output_remotes["output_remotes"].values())
        if len(ssh_remotes) == 0:
            warnings.warn("No output remotes configured in .cadet-rdm-data.json")
        for output_remote in ssh_remotes:
            try:
                print(f"Attempting to clone {output_remote} into {output_path}")
                if multi_options is None:
                    multi_options = ["--filter=blob:none"]
                _ = OutputRepo.clone(output_remote, output_path, multi_options=multi_options)
                break
            except Exception:
                traceback.print_exc()

    @staticmethod
    def _add_jupytext_file(path_root: str | Path = "."):
        jupytext_lines = ['# Pair ipynb notebooks to py:percent text notebooks', 'formats: "ipynb,py:percent"']
        write_lines_to_file(Path(path_root) / "jupytext.yml", lines=jupytext_lines, open_type="w")

    def create_remotes(self, name, namespace, url=None, username=None, push=True):
        """
        Create project in gitlab and add the projects as remotes to the project and output repositories

        :param username:
        :param url:
        :param namespace:
        :param name:
        :return:
        """
        if "github" in url:
            remote = GitHubRemote()
        else:
            remote = GitLabRemote()

        response_project = remote.create_remote(url=url, namespace=namespace, name=name, username=username)
        response_output = remote.create_remote(url=url, namespace=namespace, name=name + "_output", username=username)
        errors_encountered = 0
        try:
            self.add_remote(response_project.ssh_url_to_repo)
        except RuntimeError as e:
            errors_encountered += 1
            print(e)
            print("Please fix the error above and re-run_yml repo.add_remote()")
        try:
            self.output_repo.add_remote(response_output.ssh_url_to_repo)
        except RuntimeError as e:
            errors_encountered += 1
            print(e)
            print("Please fix the error above and re-run_yml repo.output_repo.add_remote()")
        if errors_encountered == 0 and push:
            self.push(push_all=True)

    def get_new_output_branch_name(self):
        """
        Construct a name for the new branch in the output repository.
        :return: the new branch name
        """
        project_repo_hash = str(self.head.commit)
        timestamp = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")

        if self.options and "branch_prefix" in self.options:
            branch_prefix = self.options["branch_prefix"]+"_"
        else:
            branch_prefix = ""

        branch_name = branch_prefix+"_".join([timestamp, str(self.active_branch), project_repo_hash[:7]])
        return branch_name

    def check_results_main(self):
        """
        Checkout the main branch, which contains all the log files.
        """
        self._most_recent_branch = self.output_repo.active_branch.name
        self.output_repo._git.checkout(self.output_repo.main_branch)

    def reload_recent_results(self):
        """
        Checkout the most recent previous branch.
        """
        self.output_repo._git.checkout(self._most_recent_branch)

    def print_output_log(self):
        self.output_repo.print_output_log()

    def fill_data_from_cadet_rdm_json(self, re_load=False):
        """
        Iterate through all references within the .cadet-rdm-data.json and
        load or re-load the data.

        :param re_load:
        If true: delete and re-load all data. If false, existing data will be left as-is.
        :return:
        """

        with open(self.cache_json_path, "r", encoding="utf-8") as file_handle:
            rdm_cache = json.load(file_handle)

        if "__example/path/to/repo__" in rdm_cache.keys():
            rdm_cache.pop("__example/path/to/repo__")

        for repo_location, repo_info in rdm_cache.items():
            if os.path.exists(repo_location) and re_load is False:
                continue
            elif os.path.exists(repo_location) and re_load is True:
                delete_path(repo_location)

            if repo_info["source_repo_location"] == ".":
                self.copy_data_to_cache(branch_name=repo_info["branch_name"])
            else:
                self.import_remote_repo(
                    target_repo_location=repo_location,
                    source_repo_location=repo_info["source_repo_location"],
                    source_repo_branch=repo_info["branch_name"])

    @property
    def output_log_file(self):
        # note: if filename of "log.tsv" is changed,
        #  this also has to be changed in the gitattributes of the init repo func
        return self.output_repo.output_log_file_path

    @property
    def output_log(self):
        return self.output_repo.output_log

    def _expand_tsv_header(self):
        if not self.output_log_file.exists():
            return

        with open(self.output_log_file, "r", encoding="utf-8") as f:
            lines = f.readlines()

        new_header = [
            "Output repo commit message",
            "Output repo branch",
            "Output repo commit hash",
            "Project repo commit hash",
            "Project repo folder name",
            "Project repo remotes",
            "Python sys args",
            "Tags",
            "Options hash", ]
        with open(self.output_log_file, "w", encoding="utf-8") as f:
            f.writelines(["\t".join(new_header) + "\n"])
            f.writelines(lines[1:])

        self.output_repo.add(self.output_log_file)
        self.output_repo.commit("Update tsv header", add_all=False)

    def _convert_csv_to_tsv_if_necessary(self):
        """
        If not tsv log is found AND a csv log is found, convert the csv to tsv.

        :return:
        """

        if self.output_log_file.exists():
            return

        csv_filepath = self.path / self._output_folder / "log.csv"
        if not csv_filepath.exists():
            # We have just initialized the repo and neither tsv nor csv exist.
            return

        with open(csv_filepath, encoding="utf-8") as csv_handle:
            csv_lines = csv_handle.readlines()

        tsv_lines = [line.replace(",", "\t") for line in csv_lines]

        with open(self.output_log_file, "w", encoding="utf-8") as f:
            f.writelines(tsv_lines)

        write_lines_to_file(path=self.path / ".gitattributes",
                            lines=["rdm-log.tsv merge=union"],
                            open_type="a")

    def update_output_main_logs(self, output_dict: dict = None):
        """
        Dumps all the metadata information about the project repositories state and
        the commit hash and branch name of the ouput repository into the main branch of
        the output repository.
        :param output_dict:
        Dictionary containing key-value pairs to be added to the log.
        """
        if output_dict is None:
            output_dict = {}

        output_branch_name = str(self.output_repo.active_branch)

        output_repo_hash = str(self.output_repo.head.commit)
        output_commit_message = self.output_repo.active_branch.commit.message
        output_commit_message = output_commit_message.replace("\n", "; ")

        self.output_repo._git.checkout(self.output_repo.main_branch)

        logs_folderpath = self.output_repo.path / "run_history" / output_branch_name
        if not logs_folderpath.exists():
            os.makedirs(logs_folderpath)

        json_filepath = logs_folderpath / "metadata.json"
        entry = LogEntry(
            output_repo_commit_message=output_commit_message,
            output_repo_branch=output_branch_name,
            output_repo_commit_hash=output_repo_hash,
            project_repo_commit_hash=str(self.head.commit),
            project_repo_folder_name=self.path.name,
            project_repo_remotes=self.remote_urls,
            python_sys_args=str(sys.argv),
            tags=", ".join(self.tags),
            options_hash=self.options_hash,
            filepath=None,
            **output_dict
        )

        with open(json_filepath, "w", encoding="utf-8") as f:
            json.dump(entry.to_dict(), f, indent=2)

        log = OutputLog(self.output_log_file)
        log.entries[output_branch_name] = entry
        log.write()

        self.dump_package_list(logs_folderpath)

        self._copy_code(logs_folderpath)

        self.output_repo.add(".")
        self.output_repo._git.commit("-m", f"log for '{output_commit_message}' \n"
                                           f"of branch '{output_branch_name}'")

        self.output_repo._git.checkout(output_branch_name)
        self._most_recent_branch = output_branch_name

    def _copy_code(self, target_path):
        """
        Clone only the current branch of the project repo to the target_path
        and then compress it into a zip file.

        :param target_path:
        :return:
        """
        if type(target_path) is str:
            target_path = Path(target_path)

        code_tmp_folder = target_path / "code.tar"

        self._git_repo.git.archive(
            self.active_branch, output=code_tmp_folder
        )

    def commit(self, message: str | None = None, add_all=True, verbosity=1):
        """
        Commit current state of the repository.

        :param message:
            Commit message
        :param add_all:
            Option to add all changed and new files to git automatically.
        :param verbosity:
            Option to choose degree of printed feedback.
        """
        self.update_output_remotes_json()

        super().commit(message=message, add_all=add_all, verbosity=verbosity)

    def check(self, commit=True):
        """
        Check the repository for consistency. Update remote links.

        :param commit:
        Automatically commit changes to the changed files.

        :return:
        """
        self.update_output_remotes_json()
        if commit:
            super().commit(message="Update remote links", add_all=False, verbosity=1)

        # update urls in main branch of output_repo
        self.output_repo._git.checkout(self.output_repo.main_branch)
        self.output_repo.add_list_of_remotes_in_readme_file("Link to Project Repository", self.remote_urls)
        if commit:
            self.output_repo.commit(message="Update remote links", add_all=False, verbosity=1)

    def update_output_remotes_json(self):
        output_repo_remotes = self.output_repo.remote_urls
        self.add_list_of_remotes_in_readme_file("Link to Output Repository", output_repo_remotes)

        with open(self.data_json_path, "r", encoding="utf-8") as file_handle:
            metadata = json.load(file_handle)

        remotes_dict = {remote.name: str(remote.url) for remote in self.output_repo.remotes}
        metadata["output_remotes"] = {"output_folder_name": self._output_folder, "output_remotes": remotes_dict}

        with open(self.data_json_path, "w", encoding="utf-8") as file_handle:
            json.dump(metadata, file_handle, indent=2)

        self.add(self.data_json_path)

    def download_file(self, url, file_path):
        """
        Download the file from the url and put it in the output+file_path location.

        :param file_path:
        :param url:
        :return:
            Returns a tuple containing the path to the newly created
            data file as well as the resulting HTTPMessage object.
        """
        absolute_file_path = self.output_data(file_path)
        return urlretrieve(url, absolute_file_path)

    def input_data(self, branch_name: str) -> Path:
        """
        Load previously generated results to iterate upon. Copies entire branch of output repo
        to the output_cached / branch_name folder.
        :param branch_name:
            Name of the branch of the output repository in which the results are stored.
        :return:
            Absolute path to the newly copied folder.
        """
        cached_branch_path = self.copy_data_to_cache(branch_name)

        return cached_branch_path

    @property
    def output_path(self):
        return self.output_data()

    def output_data(self, sub_path=None):
        """
        Return an absolute path with the repo_dir/output_dir/sub_path

        :param sub_path:
        :return:
        """
        if sub_path is None:
            return self.output_repo.path.absolute()
        else:
            return self.output_repo.path.absolute() / sub_path

    def remove_cached_files(self):
        """
        Delete all previously cached results.
        """
        if (self.path / (self._output_folder + "_cached")).exists():
            delete_path(self.path / (self._output_folder + "_cached"))

    def import_static_data(self, source_path: Path | str, commit_message):
        """
        Copy and commit static data from somewhere into the output repository.

        :param source_path:
        :param commit_message:
        :return:
        """
        new_branch_name = self._get_new_output_branch(force=True)
        if Path(source_path).is_dir():
            shutil.copytree(source_path, self.output_path / Path(source_path).name)
        else:
            shutil.copy(source_path, self.output_path)
        self._commit_output_data(commit_message, output_dict={})
        return new_branch_name

    def enter_context(self, force=False, debug=False):
        """
        Enter the tracking context. This includes:
         - Ensure no uncommitted changes in the project repository
         - Remove all uncommitted changes in the output repository
         - Clean up empty branches in the output repository
         - Create a new empty output branch in the output repository


        :param force:
            If False, wait for user prompts before deleting data during clean up. If True, don't wait, just delete.
        :param debug:
            If True, just return None.
        :return:
            The name of the newly created output branch.
        """
        if debug:
            return

        if self.active_branch.name == "detached_head":
            print("Repo is in detached HEAD state. Not tracking results")
            return

        self.test_for_uncommitted_changes()
        self._on_context_enter_commit_hash = self.current_commit_hash
        self._is_in_context_manager = True

        new_branch_name = self._get_new_output_branch(force)
        return new_branch_name

    def _get_new_output_branch(self, force=False):
        """
        Prepares a new branch to receive data. This includes:
         - checking out the output main branch,
         - creating a new branch from there
        This thereby produces a clear, empty directory for data, while still maintaining
        .gitignore and .gitattributes

        :param force:
            If False, wait for user prompts before deleting data during clean up. If True, don't wait, just delete.
        """

        output_repo = self.output_repo

        # ensure that LFS is properly initialized
        os.system("git lfs install")

        # Ensure clean output branch state
        if output_repo.has_uncomitted_changes:
            output_repo._reset_hard_to_head(force_entry=force)
        output_repo.delete_active_branch_if_branch_is_empty()
        new_branch_name = self.get_new_output_branch_name()

        # update urls in main branch of output_repo
        output_repo._git.checkout(output_repo.main_branch)
        project_repo_remotes = self.remote_urls
        output_repo.add_list_of_remotes_in_readme_file("Link to Project Repository", project_repo_remotes)
        output_repo.commit("Update urls", verbosity=0)

        # Create the new branch
        output_repo._git.checkout('-b', new_branch_name)  # equivalent to $ git checkout -b %branch_name
        code_backup_path = output_repo.path / "run_history"
        logs_path = output_repo.path / "log.tsv"
        if code_backup_path.exists():
            try:
                # Remove previous code backup

                delete_path(code_backup_path)
            except Exception as e:
                print(e)
        if logs_path.exists():
            try:
                # Remove previous logs
                delete_path(logs_path)
            except Exception as e:
                print(e)
        return new_branch_name

    def cache_folder_for_branch(self, branch_name=None):
        """
        Returns the path to the cache folder for the given branch

        :param branch_name:
        optional branch name, if None, current branch is used.

        :return Path:
        Path to folder in cache
        """

        branch_name_path = branch_name.replace("/", "_")

        # Define the target folder
        cache_folder = self.path / f"{self._output_folder}_cached" / str(branch_name_path)
        return cache_folder

    def copy_data_to_cache(self, branch_name=None, target_folder=None):
        """
        Copy all existing output results into a cached folder and make it read-only.

        :param branch_name:
        optional branch name, if None, current branch is used.
        :param target_folder:
        optional target directory, if None, default cache folder is used.

        :return Path:
        Path to folder in cache
        """
        # Determine the branch name if not provided
        if branch_name is None:
            branch_name = self.output_repo._git_repo.active_branch.name

        if target_folder is None:
            target_folder = self.cache_folder_for_branch(branch_name)
        target_folder = Path(target_folder)

        # Ensure that the branch is available locally. If it's only a remote branch, git.archive will fail.
        local_branches = [head.name for head in self.output_repo._git_repo.heads]
        if branch_name not in local_branches:
            self.output_repo.checkout(branch_name)

        # Create the target folder if it doesn't exist
        if not target_folder.exists():
            target_folder.mkdir(parents=True, exist_ok=True)

            # Create a temporary file for the archive and perform extraction
            handle, temp_archive_name = tempfile.mkstemp(suffix=".tar")
            os.close(handle)
            # Create an archive of the specified branch
            self.output_repo._git_repo.git.archive(
                branch_name, output=temp_archive_name
            )

            # Open the temporary file in read mode
            with open(temp_archive_name, 'rb') as archive_file:
                # Extract the archive to the specified directory
                with tarfile.open(fileobj=archive_file, mode='r', errorlevel=0, debug=1) as tar:
                    tar.extractall(path=target_folder)
            Path(temp_archive_name).unlink()

            # Set all files to read only
            for filename in glob.iglob(f"{target_folder}/**/*", recursive=True):
                absolute_path = os.path.abspath(filename)
                if os.path.isdir(absolute_path):
                    continue
                os.chmod(os.path.abspath(filename), S_IREAD)

        return target_folder

    def exit_context(self, message, output_dict: dict = None):
        """
        After running all project code, this prepares the commit of the results to the output repository. This includes
         - Ensure no uncommitted changes in the project repository
         - Stage all changes in the output repository
         - Commit all changes in the output repository with the given commit message.
         - Update the log files in the main branch of the output repository.
        :param message:
            Commit message for the output repository commit.
        :param output_dict:
            Dictionary containing optional output tracking parameters
        """
        if self._on_context_enter_commit_hash is None:
            # This means the context was not entered during enter_context
            return
        self.test_for_uncommitted_changes()
        if self._on_context_enter_commit_hash != self.current_commit_hash:
            raise RuntimeError("Code has changed since starting the context. Don't do that.")

        self._commit_output_data(message, output_dict)

    def _commit_output_data(self, message, output_dict):
        """
        Commit the data in the output repository.
         - Stage all changes in the output repository
         - Commit all changes in the output repository with the given commit message.
         - Update the log files in the main branch of the output repository.
        :param message:
            Commit message for the output repository commit.
        :param output_dict:
            Dictionary containing optional output tracking parameters
        """
        print("Completed computations, commiting results")
        self.output_repo.add(".")
        try:
            # This has to be using ._git.commit to raise an error if no results have been written.
            commit_return = self.output_repo._git.commit("-m", message)
            self.copy_data_to_cache()
            self.update_output_main_logs(output_dict)
            main_cach_path = self.path / (self._output_folder + "_cached") / self.output_repo.main_branch
            if main_cach_path.exists():
                delete_path(main_cach_path)
            self.copy_data_to_cache(self.output_repo.main_branch)
        except git.exc.GitCommandError as e:
            self.output_repo.delete_active_branch_if_branch_is_empty()
            raise e
        finally:
            # self.remove_cached_files()
            self._is_in_context_manager = False
            self._on_context_enter_commit_hash = None

    @contextlib.contextmanager
    def track_results(self, results_commit_message: str, debug=False, force=False):
        """
        Context manager to be used when running project code that produces output that should
        be tracked in the output repository.
        :param results_commit_message:
            Commit message for the commit of the output repository.
        :param debug:
            Perform calculations without tracking output.
        :param force:
            Skip confirmation and force tracking of results.
        """
        if debug:
            yield "debug"
            return

        if self.active_branch.name == "detached_head":
            print("Repo is in detached HEAD state. Not tracking results")
            yield "detached_head"
            return

        new_branch_name = self.enter_context(force=force)
        try:
            yield new_branch_name
        except Exception as e:
            self.capture_error(e)
            raise e
        else:
            self.exit_context(message=results_commit_message)

    def capture_error(self, error):
        print(traceback.format_exc())
        write_lines_to_file(self.output_path / "error.stack", traceback.format_exc().split("\n"))


class OutputRepo(BaseRepo):
    @property
    def output_log_file_path(self):
        if not self.active_branch == self.main_branch:
            self.checkout(self.main_branch)
        return self.path / "log.tsv"

    @property
    def output_log(self):
        if self.has_uncomitted_changes:
            self._reset_hard_to_head(force_entry=True)
        if not self.active_branch == self.main_branch:
            self.checkout(self.main_branch)
        return OutputLog(filepath=self.output_log_file_path)

    def update_log_hashes(self):
        if self.has_uncomitted_changes:
            self._reset_hard_to_head(force_entry=True)
        if not self.active_branch == self.main_branch:
            self.checkout(self.main_branch)
        log = OutputLog(filepath=self.output_log_file_path)

        for branch, entry in log.entries.items():
            try:
                self.checkout(branch)
            except git.GitCommandError:
                warnings.warn(f"Could not find branch {branch} in output repository.")
                continue
            if not (self.path / "options.json").exists():
                continue
            options = Options.load_json_file(self.path / "options.json")
            entry.options_hash = options.get_hash()

        self.checkout(self.main_branch)
        if self.output_log.n_entries > 0:
            log.write()
        self.commit(message="Updated log hashes", add_all=True)

    def print_output_log(self):
        self.checkout(self.main_branch)

        output_log = self.output_log
        print(output_log)

        self.checkout(self._most_recent_branch)

    def add_filetype_to_lfs(self, file_type):
        """
        Add the filetype given in file_type to the GIT-LFS tracking

        :param file_type:
        Wildcard formatted string. Examples: "*.png" or "*.xlsx"
        :return:
        """
        init_lfs(lfs_filetypes=[file_type], path=self.path)
        self.add_all_files()
        self.commit(f"Add {file_type} to lfs")


class JupyterInterfaceRepo(ProjectRepo):
    def commit(self, message: str | None = None, add_all=True, verbosity=1):
        """
        Commit current state of the repository.

        :param message:
            Commit message
        :param add_all:
            Option to add all changed and new files to git automatically.
        :param verbosity:
            Option to choose degree of printed feedback.
        """

        if "nbconvert_call" in sys.argv:
            print("Not committing during nbconvert.")
            return

        Notebook.save_ipynb()

        super().commit(message, add_all, verbosity)

    def commit_nb_output(self, notebook_path: str, results_commit_message: str,
                         force_rerun=True, timeout=600, conversion_formats: list = None):
        if "nbconvert_call" in sys.argv:
            return
            # This is reached in the first call of this function
        if not Path(notebook_path).is_absolute():
            notebook_path = self.path / notebook_path

        notebook = Notebook(notebook_path)

        with self.track_results(results_commit_message, force=True):
            notebook.check_and_rerun_notebook(force_rerun=force_rerun,
                                              timeout=timeout)

            # This is executed after the nbconvert call
            notebook.convert_ipynb(self.output_path, formats=conversion_formats)
            notebook.export_all_figures(self.output_path)



import os
from pathlib import Path
import pytest

from cadetrdm import Options, Environment, Case, ProjectRepo
from cadetrdm.container import PodmanAdapter

IMAGE_NAME = "ghcr.io/ronald-jaepel/cadet-rdm-jupyter:latest"


@pytest.mark.container
def test_run_in_podman():
    # You need to install passt on your system and add it to the path
    # os.environ["PATH"] += os.pathsep + "/home/bin/passt"
    WORK_DIR = Path.cwd() / "tmp"
    WORK_DIR.mkdir(parents=True, exist_ok=True)

    rdm_example = ProjectRepo(
        path=WORK_DIR / 'template',
        url="git@github.com:ronald-jaepel/rdm_testing_template.git",
        branch="main",
        suppress_lfs_warning=True
    )

    options = Options()
    options.debug = False
    options.push = False
    options.commit_message = 'Trying out new things'
    options.optimizer_options = {
        "optimizer": "U_NSGA3",
        "pop_size": 2,
        "n_cores": 2,
        "n_max_gen": 1,
    }

    matching_environment = Environment(
        conda_packages={
            "libsqlite": "==3.48.0"
        },
        # pip_packages={
        #     "cadet-rdm": "git+https://github.com/cadet/CADET-RDM.git@feature/container_interfaces"
        # }
    )

    case = Case(project_repo=rdm_example, options=options, environment=matching_environment)
    podman_adapter = PodmanAdapter()
    podman_adapter.image = IMAGE_NAME
    has_run_study = case.run_study(
        container_adapter=podman_adapter,
        force=True,
        # command=None  # eg: "python src/test_cadet_core/verify_cadet-core.py"
    )
    assert has_run_study

    options.optimizer_options = {
        "optimizer": "NOT_AN_OPTIMIZER",
        "pop_size": 2,
        "n_cores": 2,
        "n_max_gen": 1,
    }

    case = Case(project_repo=rdm_example, options=options, environment=matching_environment)
    podman_adapter = PodmanAdapter()
    podman_adapter.image = IMAGE_NAME
    has_run_study = case.run_study(container_adapter=podman_adapter, force=True)
    assert not has_run_study


@pytest.mark.slow
@pytest.mark.container
def test_pytest_in_podman():
    # You need to install passt on your system and add it to the path
    # os.environ["PATH"] += os.pathsep + "/home/bin/passt"
    WORK_DIR = Path.cwd() / "tmp"
    WORK_DIR.mkdir(parents=True, exist_ok=True)

    podman_adapter = PodmanAdapter()
    podman_adapter.image = IMAGE_NAME
    log, returncode = podman_adapter.run(
        "git clone git@github.com:cadet/CADET-RDM.git study && "
        "cd study && "
        "pytest tests -m 'not server_api and not docker and not slow'",

    )
    print(log)
    assert returncode == 0


@pytest.mark.container
def test_dockered_from_yml():
    podman_adapter = PodmanAdapter()
    podman_adapter.image = IMAGE_NAME
    has_run_study = podman_adapter.run_yml((Path(__file__).parent.resolve() / "case.yml").as_posix())
    assert has_run_study



import os

import git
from git import Repo

n_commits = 60

os.chdir(r"C:\Users\ronal\PycharmProjects\benchmarking-cadet-optimization\benchmarking-runner\tmp\puetmann2013\output")
repo = Repo(".")
repo.git.checkout("main")
commits = list(repo.iter_commits("main", max_count=n_commits + 1))
last_previous_commit = commits[-1]
commits = commits[:-1]

repo.git.checkout("main", b="patched_main")
repo.git.reset(last_previous_commit.hexsha, "--hard")

commit = commits[-1]

for commit in commits[::-1]:
    print(commit.hexsha)
    try:
        repo.git.cherry_pick(commit.hexsha, "-X", "theirs")
    except git.exc.GitCommandError as e:
        if "is a merge but no -m " in e.stderr:
            try:
                repo.git.cherry_pick(commit.hexsha, "-X", "theirs", "-m", "1")
            except git.exc.GitCommandError as e:
                if "The previous cherry-pick is now empty" in e.stderr:
                    pass

    status = repo.git.status()
    if "Changes not staged for commit:" in status:
        os.system("git lfs fsck --pointers & git add . & git commit --amend --no-edit")

if False:
    repo.git.checkout("main")
    repo.git.branch("-D", "patched_main")
    repo.git.branch("-D", "unpatched_main")

repo.git.checkout("main")
repo.git.reset("--hard", "patched_main")

os.system("git push --force-with-lease")

# print(repo.git.status())
# print(repo.git.log())

branches = repo.git.branch("-r").split("\n")
branches = [branch.replace("  origin/", "") for branch in branches if "origin/main" not in branch]
# remote = repo.remotes[0]


for branch in branches:
    print(branch)
    repo.git.checkout(branch)
    status = repo.git.status()
    if "Changes not staged for commit:" in status:
        os.system("git lfs fsck --pointers & git add . & git commit --amend --no-edit")

os.system("git push --all --force-with-lease")



import csv
from pathlib import Path
from typing import Dict, List

from tabulate import tabulate

from cadetrdm.environment import Environment


class LogEntry:
    def __init__(self, output_repo_commit_message, output_repo_branch, output_repo_commit_hash,
                 project_repo_commit_hash, project_repo_folder_name, project_repo_remotes, python_sys_args, tags,
                 options_hash, filepath, **kwargs):
        self.output_repo_commit_message = output_repo_commit_message
        self.output_repo_branch = output_repo_branch
        self.output_repo_commit_hash = output_repo_commit_hash
        self.project_repo_commit_hash = project_repo_commit_hash
        self.project_repo_folder_name = project_repo_folder_name
        self.project_repo_remotes = project_repo_remotes
        self.python_sys_args = python_sys_args
        self.tags = tags
        self.options_hash = options_hash
        self._filepath = filepath
        self._environment: Environment = None
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __repr__(self):
        return f"OutputEntry('{self.output_repo_commit_message}', '{self.output_repo_branch}')"

    def to_dict(self):
        return {key: value for key, value in self.__dict__.items() if not key.startswith("_")}

    @property
    def environment(self):
        if self._filepath is None:
            raise ValueError("OutputLog was initialized without a filepath, can not load Environment data.")
        if self._environment is None:
            self._load_environment()

        return self._environment

    def _load_environment(self):
        environment_path = (
                Path(self._filepath).parent
                / "run_history"
                / self.output_repo_branch
                / "conda_environment.yml"
        )
        self._environment = Environment.from_yml(environment_path)

    def matches_options_hash(self, options_hash):
        return self.options_hash == options_hash

    def matches_study_hash(self, study_hash):
        return self.project_repo_commit_hash == study_hash

    def fulfils_environment(self, environment: Environment):
        """
        Checks if this environment fulfils the requirements in a given environment.

        :param environment:
            Instance of Environment class, with requirements as key: value pairs.
        :return:
        """
        if self._environment is None:
            self._load_environment()

        return self._environment.fulfils_environment(environment)

    def package_version(self, package):
        """
        Retrieves the version of the specified package.

        Args:
            package (str): The name of the package for which the version is to be retrieved.

        Returns:
            str: The version of the specified package.
        """
        if self._environment is None:
            self._load_environment()

        return self._environment.packages[package]

    def fulfils(self, package: str, version: str):
        """
        Checks if the installed version of a package matches the specified version.

        Args:
            package (str): The name of the package to check.
            version (str): The version or specification string to match against.

        Returns:
            bool: True if the installed package version matches the specified version, False otherwise.

        Examples:
            check_package_version("conda", ">=0.1.1") -> true if larger or equal
            check_package_version("conda", "~0.1.1") -> true if approximately equal (excluding pre-release suffixes)
            check_package_version("conda", "0.1.1") -> true if exactly equal

        Uses semantic versioning to compare the versions.
        """
        if self._environment is None:
            self._load_environment()

        return self._environment.fulfils(package, version)


class OutputLog:
    def __init__(self, filepath=None):
        self._filepath = filepath

        if filepath is None or not Path(filepath).exists():
            self._entry_list = [[], []]
            self.entries = {}
            return

        self._entry_list = self._read_file(filepath)
        self.entries: Dict[str, LogEntry] = self._entries_from_entry_list(self._entry_list)

    @property
    def n_entries(self) -> int:
        """int: Number of results stored in the repository."""
        return len(self.entries)

    @classmethod
    def from_list(cls, entry_list: List[List[str]]):
        instance = cls()
        instance._entry_list = entry_list
        instance.entries: Dict[str, LogEntry] = instance._entries_from_entry_list(instance._entry_list)
        return instance

    def _entries_from_entry_list(self, entry_list) -> Dict[str, LogEntry]:
        header = self._convert_header(entry_list[0])
        if len(header) < 9:
            header.append("options_hash")
        entry_list = entry_list[1:]
        entry_dictionaries = []
        for entry in entry_list:
            if len(entry) < len(header):
                entry += [""] * (len(header) - len(entry))

            entry_dictionaries.append(
                {key: value for key, value in zip(header, entry)}
            )
        return {entry["output_repo_branch"]: LogEntry(**entry, filepath=self._filepath) for entry in entry_dictionaries}

    def _read_file(self, filepath):
        with open(filepath) as handle:
            lines = handle.readlines()
        lines = [line.replace("\n", "").split("\t") for line in lines]
        return lines

    def _convert_header(self, header):
        return [entry.lower().replace(" ", "_") for entry in header]

    def __str__(self):
        return tabulate(self._entry_list[1:], headers=self._entry_list[0])

    def __repr__(self):
        return f"OutputLog.from_list({self._entry_list})"

    @property
    def header(self):
        collection_of_keys = None
        for entry in self.entries.values():
            if collection_of_keys is None:
                collection_of_keys = entry.to_dict()
            collection_of_keys.update(entry.to_dict())

        return collection_of_keys.keys()

    def write(self):
        if self._filepath is None:
            raise ValueError("No filepath set for output log. Can not write to filepath")

        with open(self._filepath, "w", newline="") as tsv_file_handle:
            writer = csv.DictWriter(tsv_file_handle, fieldnames=self.header, delimiter="\t")
            writer.writeheader()
            for entry in self.entries.values():
                writer.writerow(entry.to_dict())



import urllib.request
import shutil
import tempfile


def prepare_conda_env(url: str = None):
    if url is None:
        url = 'https://raw.githubusercontent.com/modsim/bug_report_example/master/conda_base_environment.yml'
        print("Using default environment configuration from")
        print(url)
    with urllib.request.urlopen(url) as response:
        with tempfile.NamedTemporaryFile(delete=False, prefix="environment_", suffix=".yaml") as tmp_file:
            shutil.copyfileobj(response, tmp_file)

    print("Please now run_yml this command in a terminal (Linux) or anaconda shell (Windows):\n")
    print(f"conda deactivate && conda env create -f {tmp_file.name}\n")
    return



import os
from pathlib import Path

import pytest

from cadetrdm import Options, Study, Case, Environment, ProjectRepo, initialize_repo
from cadetrdm.io_utils import delete_path


@pytest.mark.server_api
def test_module_import():
    WORK_DIR = Path.cwd() / "tmp"
    if WORK_DIR.exists():
        delete_path(WORK_DIR)

    WORK_DIR.mkdir(parents=True, exist_ok=True)

    rdm_example = ProjectRepo(
        WORK_DIR / 'template',
        "git@jugit.fz-juelich.de:r.jaepel/rdm_example.git",
    )

    assert hasattr(rdm_example.module, "main")
    assert hasattr(rdm_example.module, "setup_optimization_problem")


@pytest.mark.server_api
def test_run_with_non_matching_env():
    WORK_DIR = Path.cwd() / "tmp"
    WORK_DIR.mkdir(parents=True, exist_ok=True)

    rdm_example = Study(
        WORK_DIR / 'template',
        "git@jugit.fz-juelich.de:r.jaepel/rdm_example.git",
    )

    options = Options()
    options.debug = True
    options.push = False
    options.commit_message = 'Trying out new things'
    options.optimizer_options = {
        "optimizer": "U_NSGA3",
        "pop_size": 2,
        "n_cores": 2,
        "n_max_gen": 1,
    }

    matching_environment = Environment(conda_packages={
            "cadet": ">1.0.0"
        }
    )

    case = Case(project_repo=rdm_example, options=options, environment=matching_environment)
    assert case.can_run_study is True

    non_matching_environment = Environment(conda_packages={
            "cadet": "17.0.0"
        }
    )

    case = Case(project_repo=rdm_example, options=options, environment=non_matching_environment)
    assert case.can_run_study is False


@pytest.mark.server_api
def test_re_load_results():
    from pathlib import Path

    from cadetrdm import Options
    from cadetrdm import Study, Case

    WORK_DIR = Path("batch_repos") / "studies"

    batch_elution = Study(
        WORK_DIR / 'batch_elution',
        "git@jugit.fz-juelich.de:j.schmoelder/batch_elution.git",
        branch="master",
    )

    options = Options({
        "_cadet_options": {"install_path": None, "use_dll": True},
        "_temp_directory_base": {"class": "Path", "value": "/dev/shm/schmoelder/CADET-Process/tmp"},
        "_cache_directory_base": {"class": "Path", "value": "/dev/shm/schmoelder/CADET-Process/cache"},
        "objective": "single-objective",
        "optimizer_options": {"optimizer": "U_NSGA3", "n_cores": -4, "n_max_gen": 64},
        "commit_message": "single-objective_2025-01-20",
        "debug": False,
        "push": True
    })

    single_objective_case = Case(batch_elution, options, name=options.objective)
    results_path = single_objective_case.results_path
    assert results_path.exists()

@pytest.mark.server_api
def test_results_loading():
    WORK_DIR = Path.cwd() / "tmp"
    WORK_DIR.mkdir(parents=True, exist_ok=True)

    rdm_example = Study(
        WORK_DIR / 'template',
        "git@jugit.fz-juelich.de:r.jaepel/rdm_example.git",
    )

    class OptionsFixture(Options):
        def get_hash(self):
            return "za16jkxf3waxxy3mkavy3jnjn5za0b"

    case = Case(project_repo=rdm_example, options=OptionsFixture())
    assert case.has_results_for_this_run
    assert case.results_branch == '2025-02-11_17-15-38_main_d1842fd'

    simple_environment = Environment(
        conda_packages={"cadet": "4.4.0",
                        "mkl": ">=2024",
                        "tbb": ">2021.10"},
        pip_packages={'about-time': '4.2.1'}
    )
    mismatched_environment = Environment(
        conda_packages={"cadet": "4.5.0",
                        "mkl": ">2024",
                        "tbb": "<2021.10"},
        pip_packages={'about-time': '4.3.1'}
    )
    full_environment = Environment(
        conda_packages={'bzip2': '1.0.8', 'ca-certificates': '2024.2.2', 'cadet': '4.4.0', 'hdf5': '1.14.3',
                        'intel-openmp': '2024.0.0', 'krb5': '1.21.2', 'libaec': '1.1.2', 'libblas': '3.9.0',
                        'libcblas': '3.9.0', 'libcurl': '8.5.0', 'libexpat': '2.6.1', 'libffi': '3.4.2',
                        'libhwloc': '2.9.3', 'libiconv': '1.17', 'liblapack': '3.9.0', 'liblapacke': '3.9.0',
                        'libsqlite': '3.45.2', 'libssh2': '1.11.0', 'libxml2': '2.12.5', 'libzlib': '1.2.13',
                        'mkl': '2024.0.0', 'openssl': '3.2.1', 'pip': '24.0', 'pthreads-win32': '2.9.1',
                        'python': '3.11.8', 'suitesparse': '5.4.0', 'tbb': '2021.11.0', 'tbb-devel': '2021.11.0',
                        'tk': '8.6.13', 'ucrt': '10.0.22621.0', 'vc': '14.3', 'vc14_runtime': '14.38.33130',
                        'vs2015_runtime': '14.38.33130', 'wheel': '0.42.0', 'xz': '5.2.6'},
        pip_packages={'about-time': '4.2.1', 'addict': '2.3.0', 'alive-progress': '3.1.5', 'annotated-types': '0.6.0',
                      'anyio': '4.3.0', 'appdirs': '1.4.4', 'arviz': '0.17.0', 'attrs': '23.2.0', 'autograd': '1.6.2',
                      'betterproto': '2.0.0b6', 'build': '1.1.1', 'cadet-process': '0.8.0', 'cadet-python': '0.14.1',
                      'cadet-rdm': '0.0.32', 'certifi': '2024.2.2', 'cffi': '1.16.0', 'cfgv': '3.4.0',
                      'charset-normalizer': '3.3.2', 'click': '8.1.7', 'cma': '3.2.2', 'cobra': '0.29.0',
                      'colorama': '0.4.6', 'contourpy': '1.2.0', 'corner': '2.2.2', 'cryptography': '42.0.5',
                      'cycler': '0.12.1', 'depinfo': '2.2.0', 'deprecated': '1.2.14', 'dill': '0.3.8',
                      'diskcache': '5.6.3', 'distlib': '0.3.8', 'filelock': '3.13.1', 'flake8': '7.0.0',
                      'fonttools': '4.49.0', 'future': '1.0.0', 'git-lfs': '1.6', 'gitdb': '4.0.11',
                      'gitpython': '3.1.42', 'grapheme': '0.6.0', 'grpclib': '0.4.7', 'h11': '0.14.0', 'h2': '4.1.0',
                      'h5netcdf': '1.3.0', 'h5py': '3.10.0', 'hagelkorn': '1.2.3', 'hopsy': '1.4.1', 'hpack': '4.0.0',
                      'httpcore': '1.0.4', 'httpx': '0.27.0', 'hyperframe': '6.0.1', 'identify': '2.5.35',
                      'idna': '3.6', 'importlib-metadata': '7.0.2', 'importlib-resources': '6.3.0',
                      'iniconfig': '2.0.0', 'jaraco-classes': '3.3.1', 'joblib': '1.3.2', 'keyring': '24.3.1',
                      'kiwisolver': '1.4.5', 'llvmlite': '0.42.0', 'markdown-it-py': '3.0.0', 'matplotlib': '3.8.3',
                      'mcbackend': '0.5.2', 'mccabe': '0.7.0', 'mdurl': '0.1.2', 'more-itertools': '10.2.0',
                      'mpmath': '1.3.0', 'multidict': '6.0.5', 'multiprocess': '0.70.16', 'nodeenv': '1.8.0',
                      'numba': '0.59.0', 'numpy': '1.26.4', 'optlang': '1.8.1', 'packaging': '24.0', 'pandas': '2.2.1',
                      'pathos': '0.3.2', 'pillow': '10.2.0', 'platformdirs': '4.2.0', 'pluggy': '1.4.0',
                      'polyround': '0.2.11', 'pox': '0.3.4', 'ppft': '1.7.6.8', 'pre-commit': '3.6.2',
                      'pycodestyle': '2.11.1', 'pycparser': '2.21', 'pydantic': '2.6.4', 'pydantic-core': '2.16.3',
                      'pyflakes': '3.2.0', 'pygithub': '2.2.0', 'pygments': '2.17.2', 'pyjwt': '2.8.0',
                      'pymoo': '0.6.1.1', 'pynacl': '1.5.0', 'pyparsing': '3.1.2', 'pyproject-hooks': '1.0.0',
                      'pyqt5': '5.15.10', 'pyqt5-qt5': '5.15.2', 'pyqt5-sip': '12.13.0', 'pytest': '8.1.1',
                      'python-benedict': '0.33.2', 'python-dateutil': '2.9.0.post0', 'python-fsutil': '0.13.1',
                      'python-gitlab': '4.4.0', 'python-libsbml': '5.20.2', 'python-slugify': '8.0.4', 'pytz': '2024.1',
                      'pywin32-ctypes': '0.2.2', 'pyyaml': '6.0.1', 'requests': '2.31.0', 'requests-toolbelt': '1.0.0',
                      'rich': '13.7.1', 'ruamel-yaml': '0.18.6', 'ruamel-yaml-clib': '0.2.8',
                      'scikit-learn': '1.4.1.post1', 'scipy': '1.12.0', 'setuptools': '69.2.0', 'six': '1.16.0',
                      'smmap': '5.0.1', 'sniffio': '1.3.1', 'swiglpk': '5.0.10', 'sympy': '1.12', 'tabulate': '0.9.0',
                      'text-unidecode': '1.3', 'threadpoolctl': '3.3.0', 'tqdm': '4.66.2',
                      'typing-extensions': '4.10.0', 'tzdata': '2024.1', 'urllib3': '2.2.1', 'virtualenv': '20.25.1',
                      'wrapt': '1.16.0', 'xarray': '2024.2.0', 'xarray-einstats': '0.7.0', 'zipp': '3.18.0'}
    )

    case = Case(project_repo=rdm_example, options=OptionsFixture(), environment=simple_environment)
    assert case.has_results_for_this_run
    case = Case(project_repo=rdm_example, options=OptionsFixture(), environment=full_environment)
    assert case.has_results_for_this_run
    case = Case(project_repo=rdm_example, options=OptionsFixture(), environment=mismatched_environment)
    assert not case.has_results_for_this_run

    # delete_path(WORK_DIR)


def test_case_with_projectrepo():
    class OptionsFixture(Options):
        def get_hash(self):
            return "za16jkxf3waxxy3mkavy3jnjn5za0b"

    path_to_repo = Path("test_repo_batch")
    if path_to_repo.exists():
        delete_path(path_to_repo)

    initialize_repo(path_to_repo)

    try:
        os.chdir(path_to_repo)
        Case(project_repo=ProjectRepo(), options=OptionsFixture())

    finally:
        os.chdir("..")

    return


@pytest.mark.server_api
def test_results_loading_from_within():
    class OptionsFixture(Options):
        def get_hash(self):
            return "za16jkxf3waxxy3mkavy3jnjn5za0b"

    root_dir = os.getcwd()

    WORK_DIR = Path.cwd() / "tmp"
    WORK_DIR.mkdir(parents=True, exist_ok=True)

    if (WORK_DIR / 'template').exists():
        delete_path(WORK_DIR / 'template')

    rdm_example = Study(
        WORK_DIR / 'template',
        "git@jugit.fz-juelich.de:r.jaepel/rdm_example.git",
    )

    try:
        os.chdir(WORK_DIR / 'template')
        rdm_example = Study(".", "git@jugit.fz-juelich.de:r.jaepel/rdm_example.git")

        case = Case(project_repo=rdm_example, options=OptionsFixture())
        assert case.has_results_for_this_run
        assert case.results_branch == '2025-02-11_17-15-38_main_d1842fd'

        with open(rdm_example.path / "Readme.md", "a") as handle:
            handle.write("New line\n")
        rdm_example.commit("modify readme")

        case = Case(project_repo=rdm_example, options=OptionsFixture())
        assert not case.has_results_for_this_run
    finally:
        os.chdir(root_dir)


if __name__ == "__main__":
    test_results_loading_from_within()



import io
import re
from typing import Self, List
from typing import Dict as DictType
import subprocess

import yaml
from semantic_version import Version, SimpleSpec


class Environment:

    def __init__(self, *args, conda_packages: DictType[str, str] = None, pip_packages: DictType[str, str] = None,
                 name: str = None, channels: List[str] = None):
        if args:
            raise TypeError(
                "Environment.__init__() does not take positional arguments. "
                "Please specify conda_packages and/or pip_packages."
            )
        self.conda_packages = conda_packages
        self.pip_packages = pip_packages
        self.packages = dict()
        if conda_packages:
            self.packages.update(conda_packages)
        if pip_packages:
            self.packages.update(pip_packages)

        self.name = name
        self.channels = channels

    @classmethod
    def from_yml(cls, yml_path):
        """
        Create an Environment object from a YAML file.

        :param yml_path:
        :return:
        """
        with open(yml_path) as handle:
            yml_string = "".join(handle.readlines())

        instance = cls.from_yml_string(yml_string)
        return instance

    @classmethod
    def from_yml_string(cls, yml_string):
        """
        Creates an Environment object from a YAML string.

        :param yml_string:
        :return:
        """

        # Remove special formatting characters from the string
        ansi_escape_pattern = re.compile(r'\x1b\[[0-9;]*[a-zA-Z]')
        yml_string = re.sub(ansi_escape_pattern, "", yml_string)

        packages = yaml.safe_load(yml_string)

        instance = cls()

        # If for some reason the environment.yml is empty, return an empty instance
        if packages is None:
            return instance

        instance.name = packages["name"]
        instance.channels = packages["channels"]

        conda_packages = packages["dependencies"]
        conda_packages = {line.split("=")[0]: line.split("=")[1] for line in conda_packages if isinstance(line, str)}
        instance.packages.update(conda_packages)
        instance.conda_packages = conda_packages

        if isinstance(packages["dependencies"][-1], dict):
            pip_packages = packages["dependencies"][-1]["pip"]
            pip_packages = {line.split("==")[0]: line.split("==")[1] for line in pip_packages}
            instance.packages.update(pip_packages)
            instance.pip_packages = pip_packages

        return instance

    def to_yml(self, handle):
        """
        Create an environment.yml file from an Environment instance.

        :param handle:
        :return:
        """
        yml_dict = self._to_yml_dict()

        yaml.safe_dump(yml_dict, handle)

    def _to_yml_dict(self):
        """
        Create an environment.yml type yml dict from an Environment instance.

        :return: yml dict
        """
        dependency_list = []
        if self.conda_packages is not None:
            for package, spec in self.conda_packages.items():
                if "git+" in spec:
                    raise ValueError(f"Conda can not use git+ dependencies for {package} {spec}")
                elif ">" in spec or "<" in spec or "=" in spec:
                    dependency_list.append(f"{package}{spec}")
                else:
                    dependency_list.append(f"{package}={spec}")

        pip_list = []
        if self.pip_packages is not None:
            for package, spec in self.pip_packages.items():
                if "git+" in spec:
                    pip_list.append(spec)
                elif ">" in spec or "<" in spec or "=" in spec:
                    pip_list.append(f"{package}{spec}")
                else:
                    pip_list.append(f"{package}=={spec}")

            dependency_list.append({"pip": pip_list})

        yml_dict = {
            "name": self.name,
            "channels": self.channels,
            "dependencies": dependency_list,
        }
        return yml_dict

    def update(self, environment: Self):
        if environment.name is not None:
            self.name = environment.name
        if environment.channels is not None:
            self.channels = environment.channels
        self.conda_packages.update(environment.conda_packages)
        self.pip_packages.update(environment.pip_packages)

    def package_version(self, package):
        if package not in self.packages:
            return None

        return self.packages[package]

    def fulfils(self, package, version):
        """
        Checks if the installed version of a package matches the specified version.

        Args:
            package (str): The name of the package to check.
            version (str): The version or specification string to match against.

        Returns:
            bool: True if the installed package version matches the specified version, False otherwise.

        Examples:
            check_package_version("conda", ">=0.1.1") -> true if larger or equal
            check_package_version("conda", "~0.1.1") -> true if approximately equal (tolerant of pre-release suffixes)
            check_package_version("conda", "0.1.1") -> true if exactly equal (must match pre-release suffixes)

        Uses semantic versioning to compare the versions.
        """

        installed_version = self.package_version(package)
        if installed_version is None:
            return False

        if "git+" in installed_version:
            return False

        # Use .coerce instead of .parse to ensure non-standard version strings are converted.
        # Rules are:
        #   - If no minor or patch component, and partial is False, replace them with zeroes
        #   - Any character outside of a-zA-Z0-9.+- is replaced with a -
        #   - If more than 3 dot-separated numerical components,
        #       everything from the fourth component belongs to the build part
        #   - Any extra + in the build part will be replaced with dots
        installed_version = Version.coerce(installed_version)

        try:
            spec = SimpleSpec(version)
        except ValueError as e:
            spec = SimpleSpec(str(Version.coerce(version)))
            print(f"Warning: {e} when processing {package}={version}. Using {str(Version.coerce(version))} instead.")

        match = spec.match(installed_version)

        return match

    def fulfils_environment(self, environment: Self):
        """
        Checks if this environment fulfils the requirements in a given environment.

        :param environment:
            Instance of Environment class, with requirements as key: value pairs.
        :return:
        """

        if environment is None:
            return True

        mismatches = []

        for package, version in environment.packages.items():
            try:
                if not self.fulfils(package, version):
                    mismatches.append((package, version, self.package_version(package)))
            except ValueError:
                mismatches.append((package, version, self.package_version(package)))

        if mismatches:
            for package, version, existing_version in mismatches:
                print(f"Package {package}: {existing_version} does not fulfil requirements: {version}")
            return False

        return True

    def prepare_install_instructions(self):

        conda_command = ""
        pip_command = ""

        if self.conda_packages is not None:
            conda_command = "conda install -y"
            for package, spec in self.conda_packages.items():
                if "git+" in spec:
                    raise ValueError(f"Conda can not use git+ dependencies for {package} {spec}")
                elif "~" in spec:
                    conda_command += f" {package}={spec.replace('~', '').replace('=', '')}"
                elif ">" in spec or "<" in spec or "=" in spec:
                    conda_command += f" {package}{spec}"
                else:
                    conda_command += f" {package}={spec}"

        if self.pip_packages is not None:
            pip_command = "pip install"
            pip_reinstall_command = "pip install --force-reinstall --no-deps"
            reinstall_necessary = False
            for package, spec in self.pip_packages.items():
                if "git+" in spec:
                    pip_command += f" '{spec}'"
                    pip_reinstall_command += f" '{spec}' "
                    reinstall_necessary = True
                elif "~" in spec:
                    pip_command += f" '{package}={spec.replace('~', '').replace('=', '')}'"
                elif ">" in spec or "<" in spec or "=" in spec:
                    pip_command += f" '{package}{spec}'"
                else:
                    pip_command += f" '{package}=={spec}'"
            if reinstall_necessary:
                pip_command = pip_command + " && " + pip_reinstall_command

        if conda_command and pip_command:
            install_instructions = conda_command + " && " + pip_command
            return install_instructions
        elif conda_command:
            return conda_command
        elif pip_command:
            return pip_command
        else:
            return None

    def __repr__(self):
        return (f"Environment("
                f"conda_packages = {repr(self.conda_packages)},  "
                f"pip_packages = {repr(self.pip_packages)}"
                f")")

    def __str__(self):
        handle = io.StringIO()
        yaml.safe_dump(self._to_yml_dict(), handle)
        return "Environment:\n" + handle.getvalue()



import hashlib
import json
from pathlib import Path

from addict import Dict
import numpy as np


def remove_invalid_keys(dicti, excluded_keys=None):
    if excluded_keys is None:
        excluded_keys = []

    def is_valid(key):
        return not (key.startswith("_") or "__" in key or key in excluded_keys)

    new_dicti = {}
    for key, value in dicti.items():
        if not is_valid(key):
            continue
        if isinstance(value, dict):
            value = remove_invalid_keys(value)
        new_dicti[key] = value

    return new_dicti


class CustomEncoder(json.JSONEncoder):
    """Custom encoder to serialize additional types (e.g. numpy arrays) to json."""

    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return {"__class__": "numpy.ndarray", "value": obj.tolist()}
        elif isinstance(obj, Path):
            return {"__class__": "Path", "value": obj.as_posix()}
        return json.JSONEncoder.default(self, obj)


class CustomDecoder(json.JSONDecoder):
    """Custom decoder to deserialize additional types (e.g. numpy arrays) from json."""

    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)

    def object_hook(self, obj):
        import numpy
        if '__class__' not in obj:
            return obj
        match obj['__class__']:
            case 'numpy.ndarray':
                return numpy.array(obj['value'])
            case 'Path':
                return Path(obj['value'])
        return obj


class Options(Dict):
    def dumps(self):
        return json.dumps(dict(self), cls=CustomEncoder)

    def copy(self):
        new = super().copy()
        return Options(new)

    # super.update() already takes care of nested dictionaries, so we don't have to

    @classmethod
    def loads(cls, string):
        decoded = json.loads(string, cls=CustomDecoder)
        return cls(decoded)

    @classmethod
    def load_json_file(cls, file_path, **loader_kwargs):
        with open(file_path, "r") as handle:
            json_data = json.load(handle, cls=CustomDecoder, **loader_kwargs)
        return cls(json_data)

    def dump_json_file(self, file_path, **dumper_kwargs):
        with open(file_path, "w") as handle:
            json.dump(dict(self), handle, cls=CustomEncoder, **dumper_kwargs)

    def dump_json_str(self, **dumper_kwargs):
        return self.dumps()

    @classmethod
    def load_json_str(cls, string, **loader_kwargs):
        return cls.loads(string)

    def get_hash(self):
        excluded_keys = {"commit_message", "push", "debug", "force"}
        remaining_dict = remove_invalid_keys(self, excluded_keys=excluded_keys)
        dump = json.dumps(
            remaining_dict,
            cls=CustomEncoder,
            ensure_ascii=False,
            sort_keys=True,
            indent=None,
            separators=(',', ':'),
        )

        hash_alphabet = "abcdefghjkmnpqrstvwxyz0123456789"
        hash_base = len(hash_alphabet)

        def to_base(number, base):
            result = ""
            while number:
                result += hash_alphabet[number % base]
                number //= base
            return result[::-1] or "0"

        base_16_hash = hashlib.sha1(dump.encode('utf-8')).hexdigest()
        base_10_hash = int(base_16_hash, 16)
        base_32_hash = to_base(base_10_hash, hash_base)

        return base_32_hash

    def __eq__(self, other):
        if not isinstance(other, Options):
            try:
                other = Options(other)
            except TypeError:
                print(f"TypeError when casting {other} to Options()")
                return NotImplemented

        return self.get_hash() == other.get_hash()


if __name__ == '__main__':
    options = Options()
    options.optimizer_options = 10
    options.commit_message = "Fuubar"
    options_rev = Options.load_json_str(options.dump_json_str())
    print(options.dump_json_str())
    options_rev.commit_message = "unfoo"
    print(options.__hash__(), options_rev.__hash__())



import sys
import time
import os
import traceback
from pathlib import Path
from cadetrdm.io_utils import wait_for_user


class Notebook:
    def __init__(self, notebook_path):

        import nbformat as nbf
        self.nbf = nbf

        try:
            from nbconvert.preprocessors import ExecutePreprocessor
            from nbconvert.nbconvertapp import NbConvertApp
            self.ExecutePreprocessor = ExecutePreprocessor
            self.NbConvertApp = NbConvertApp
        except ModuleNotFoundError as e:
            traceback.print_exc()
            print("No working nbconvert installation found OR a conflict in your packages found.")
            print("For more information, import nbconvert and check the error.")

        self.notebook_path = Path(notebook_path)

    @property
    def notebook_name(self):
        return str(self.notebook_path.name).replace(".", "_")

    def check_execution_order(self,
                              check_all_executed=False,
                              check_top_to_bottom=False,
                              check_in_order=True,
                              exclude_last_cell=False):
        notebook = self.nbf.read(self.notebook_path, self.nbf.NO_CONVERT)

        # extract all code cells (disregard markdown, raw and others), then extract the execution order
        output_cells = [cell for cell in notebook.cells if cell["cell_type"] == "code"]
        # remove empty cells
        non_empty_cells = [cell for cell in output_cells if cell["source"] != ""]
        execution_counts = [cell["execution_count"] for cell in non_empty_cells]

        def _all_none(item_list):
            return all([i is None for i in item_list])

        # return early if no cells were executed
        if _all_none(execution_counts):
            return True

        pass_check = [True]

        def _check_all_executed(execution_counts: list) -> bool:
            """Check all cells were executed.

            Parameters
            ----------
            execution_counts : list
                execution_counts

            Returns
            -------
            bool
            """
            return not None in execution_counts

        def _check_in_order(execution_counts: list) -> bool:
            """Check that execution counts that aren't None go from 1 to N.

            Parameters
            ----------
            execution_counts : list
                execution counts

            Returns
            -------
            bool
            """
            execution_counts = [x for x in execution_counts if x is not None]
            count_range = len(execution_counts) - 1
            if exclude_last_cell:
                count_range = count_range - 1
            print(execution_counts)
            is_in_order = all([execution_counts[i] < execution_counts[i + 1] for i in range(count_range)])
            return is_in_order

        if check_in_order:
            pass_check.append(_check_in_order(execution_counts))

        if check_all_executed:
            pass_check.append(_check_all_executed(execution_counts))

        if check_top_to_bottom:
            pass_check.append(
                _check_all_executed(execution_counts) and _check_in_order(execution_counts)
            )
        return all(pass_check)

    @staticmethod
    def save_ipynb():
        from ipylab import JupyterFrontEnd
        app = JupyterFrontEnd()
        print("Saving", end="")
        # note: docmanager:save doesn't lock the python thread until saving is completed.
        # Sometimes, new changes aren't completely saved before checks are performed.
        # Waiting for 0.1 seconds seems to prevent that.
        app.commands.execute('docmanager:save')
        time.sleep(0.1)
        print("")

    def reload_notebook(self):
        from ipylab import JupyterFrontEnd
        app = JupyterFrontEnd()
        app.commands.execute('docmanager:reload')

    def check_and_rerun_notebook(self, force_rerun=False, timeout=600):
        if "nbconvert_call" in sys.argv:
            return

        self.save_ipynb()
        # wait a second for the save process to finish.
        time.sleep(1)

        is_in_order = self.check_execution_order(exclude_last_cell=False)

        if not force_rerun:
            if is_in_order:
                print("Notebook was already executed in order.")
                return
            else:
                rerun_confirmed_bool = wait_for_user("Rerun notebook now?")
                if not rerun_confirmed_bool:
                    print("Aborting.")
                    return

        print("Rerunning.")
        with open(self.notebook_path) as f:
            nb = self.nbf.read(f, as_version=4)

        ep = self.ExecutePreprocessor(timeout=timeout, kernel_name='python3', extra_arguments=["nbconvert_call"])
        ep.preprocess(nb, )

        with open(self.notebook_path, 'w', encoding='utf-8') as f:
            self.nbf.write(nb, f)

        self.reload_notebook()

    def convert_ipynb(self, output_dir, formats: list = None):
        if formats is None:
            formats = ["html", "ipynb"]
        app = self.NbConvertApp()
        app.initialize()
        output_root_directory = os.path.join(output_dir, self.notebook_name)
        for export_format in formats:
            app.export_format = export_format
            app.notebooks = [str(self.notebook_path)]
            app.output_base = os.path.join(output_root_directory,
                                           self.notebook_path.name.replace('.ipynb', ''))
            if not os.path.exists(output_root_directory):
                os.makedirs(output_root_directory)
            app.start()

    def export_all_figures(self, output_dir):
        import junix

        file_without_extension = self.notebook_path.stem
        images = junix.export_images(filepath=str(self.notebook_path),
                                     output_dir=os.path.join(output_dir, self.notebook_name),
                                     prefix=file_without_extension)
